from datetime import datetime

TYPES = [
    "number",
    "int",
    "uint",
    "string",
    "bool",
    "enum",
    "bytes",
    "map",
    "array",
    "dynamic"
]


class Value:
    def __init__(self):
        """
        Constructor for Value.
        """
        self.value = None
        self.type = None
        self.updated_at = None

    def set_type(self, t):
        """
        Set the type for this Value.
        :param t: Value type.
        :return: True on success.
        """
        if self.is_enum(t):
            pass
        elif t not in TYPES:
            raise TypeError("%s is not an acceptable type" % t)
        self.type = t

    def set_value(self, value, check=True):
        """
        Set the value.
        :param value: Value to set.
        :param check: Set to false to skip type checking.
        :return: True if successful.
        """
        set_val = True
        if check:
            set_val = self.check_type(value)
        if set_val:
            self.value = value
            self.updated_at = datetime.now()
            return True
        return False

    def has_value(self):
        return self.value is not None and self.type is not None

    def check_type(self, value):
        """
        Check the type of a variable compared to the type of this Value.
        :param value: Variable to check.
        :return: True if matches.
        """
        if self.type == "string":
            return isinstance(value, basestring)
        elif self.type == "number":
            return type(value) == int or type(value) == long or type(value) == float
        elif self.type == "int":
            return type(value) == int or type(value) == long
        elif self.type == "uint":
            return type(value) == int or type(value) == long and value > 0
        elif self.type == "bool":
            # TODO(logangorence): Implement enum-like bool. Examples: "bool[disabled,enabled]" or "bool[on,off]"
            return type(value) == bool
        elif self.is_enum(self.type):
            return value in self.get_enum_values(self.type)
        # TODO(logangorence): Bytes value type.
        elif self.type == "map":
            return type(value) == map
        elif self.type == "array":
            return type(value) == list
        elif self.type == "dynamic":
            # TODO(logangorence): Check to ensure that the type is still a valid one that we accept.
            return True

    @staticmethod
    def build_enum(values):
        """
        Utility to convert list to String
        :param values: Values for conversion
        :return: Converted string. Example: enum[foo,bar]
        """
        if type(values) == list:
            return "enum[" + ",".join(values) + "]"
        # TODO(logangorence) Add support for Python Enum class.
        else:
            raise KeyError("build_enum called with non-list parameter.")

    @staticmethod
    def get_enum_values(enum):
        if enum.startswith("enum[") and enum.endswith("]"):
            return enum[5:-1].split(",")
        else:
            raise ValueError("Not an enum!")

    @staticmethod
    def is_enum(enum_type):
        if enum_type.startswith("enum[") and enum_type.endswith("]"):
            i = enum_type[5:-1]
            # noinspection PyBroadException
            try:
                i.split(",")
                return True
            except:
                return False
        return False



import base64
import json
import time
import requests


class Handshake:
    """
    Class that handles the DSA Handshake.
    """

    def __init__(self, link, keypair):
        self.link = link
        self.keypair = keypair

    def get_dsid(self):
        return self.link.config.name + "-" + self.keypair.b64

    def get_publickey(self):
        return self.keypair.encoded_public

    def get_json(self):
        return json.dumps({
            "publicKey": self.get_publickey(),
            "isRequester": self.link.config.requester,
            "isResponder": self.link.config.responder,
            "version": "1.1.1"
        }, sort_keys=True)

    def run_handshake(self):
        url = self.link.config.broker + "?dsId=%s" % self.get_dsid()
        token = self.link.config.token_hash()
        if token is not None:
            url += token
        self.link.logger.debug("Running handshake on %s" % url)
        keep_trying = True
        cooldown = 1
        while keep_trying:
            try:
                response = requests.post(url, data=self.get_json())
                if response.status_code is 200:
                    keep_trying = False
                    continue
            except requests.exceptions.ConnectionError:
                pass
            self.link.logger.info("Failed to connect to %s" % url)
            time.sleep(cooldown)
            if cooldown <= 60:
                cooldown += 1
        # noinspection PyUnboundLocalVariable
        self.link.server_config = json.loads(response.text)
        if "tempKey" in self.link.server_config:
            self.link.needs_auth = True
            self.link.shared_secret = self.keypair.keypair.get_ecdh_key(
                base64.urlsafe_b64decode(self.link.add_padding(self.link.server_config["tempKey"]).encode("utf-8")))



from dslink.DSLink import DSLink, Configuration
from dslink.Node import Node
from dslink.Permission import Permission
from dslink.Profile import Profile
from dslink.Requester import Requester
from dslink.Response import *
from dslink.Value import Value

__title__ = "dslink"
__version__ = "0.6.11"
__author__ = "Logan Gorence"

__license__ = "Apache 2.0"
__copyright__ = "Copyright 2015 DGLogik Inc."



import json
import logging

from autobahn.twisted.websocket import WebSocketClientProtocol, WebSocketClientFactory
from autobahn.websocket.protocol import parseWsUrl
from twisted.internet import reactor
from twisted.internet.protocol import ReconnectingClientFactory

from dslink.Request import Request
from dslink.Response import Response


class WebSocket:
    """
    Class to handle WebSocket.
    """

    def __init__(self, link):
        """
        WebSocket Constructor.
        :param link: DSLink instance.
        """
        websocket_uri, url, port = link.get_url()

        DSAWebSocket.link = link
        factory = DSAWebSocketFactory(websocket_uri, link)
        factory.protocol = DSAWebSocket

        link.logger.debug("Connecting WebSocket to %s" % websocket_uri)
        reactor.connectTCP(url.hostname, port, factory)


class DSAWebSocketFactory(WebSocketClientFactory, ReconnectingClientFactory):
    def __init__(self, websocket_uri, link):
        super(DSAWebSocketFactory, self).__init__(websocket_uri)
        self.link = link
    
    def clientConnectionFailed(self, connector, reason):
        print("Failed to connect, retrying...")
        self.link.handshake.run_handshake()
        self.reset_url()
        self.retry(connector)

    def clientConnectionLost(self, connector, unused_reason):
        print("Connection lost, retrying...")
        self.link.handshake.run_handshake()
        self.reset_url()
        self.retry(connector)

    def reset_url(self):
        websocket_uri, url, port = self.link.get_url()
        self.url = websocket_uri
        (self.isSecure, self.host, self.port, self.resource, self.path, self.params) = parseWsUrl(websocket_uri)


class DSAWebSocket(WebSocketClientProtocol):
    """
    Autobahn implementation for DSA communications.
    """
    def __init__(self):
        """
            Constructor for DSAWebSocket.
            """
        super(DSAWebSocket, self).__init__()
        self.msg = 0
        self.logger = logging.getLogger("DSLink")
        self.link = DSAWebSocket.link
        self.link.wsp = self

    def sendPingMsg(self):
        """
        Send a blank object for a ping.
        """
        self.logger.debug("Sent ping")
        # noinspection PyTypeChecker
        self.sendMessage({})
        reactor.callLater(self.link.config.ping_time, self.sendPingMsg)

    def onOpen(self):
        """
        WebSocket open event.
        """
        self.link.active = True
        self.logger.info("WebSocket Connection Established")
        self.sendPingMsg()

    def onClose(self, wasClean, code, reason):
        """
        WebSocket close event.
        :param wasClean: True if the close was clean.
        :param code: Close code.
        :param reason: Close reason.
        """
        self.link.active = False
        self.logger.info("WebSocket Connection Lost")

    def onMessage(self, payload, isBinary):
        """
        WebSocket message event.
        :param payload: Data to send.
        :param isBinary: True if the message is binary.
        """
        self.logger.debug("Received data: %s" % payload.decode("utf-8"))
        i = json.loads(payload.decode("utf-8"))
        m = {}
        ack = False
        if "requests" in i and len(i["requests"]) > 0:
            ack = True
            m["responses"] = self.handleRequests(i["requests"])
        if "responses" in i and len(i["responses"]) > 0:
            ack = True
            self.handleResponses(i["responses"])
        if ack:
            if "msg" in i:
                m["ack"] = i["msg"]
            # noinspection PyTypeChecker
            self.sendMessage(m)

    def handleRequests(self, requests):
        """
        Handle incoming requests.
        :param requests: List of requests.
        :return: Outgoing responses.
        """
        i = []
        for request in requests:
            i.append(Request(request, self.link).process().get_stream())
        return i

    def handleResponses(self, responses):
        """
        Handle incoming responses.
        :param responses: List of responses.
        """
        for response in responses:
            Response(response)
            if response["rid"] is 0:
                for update in response["updates"]:
                    sid = update[0]
                    val = update[1]
                    time = update[2]
                    self.link.requester.subscription_manager.run_callback(sid, val, time)
            elif response["rid"] in self.link.requester.request_manager.requests:
                self.link.requester.request_manager.invoke_request(response["rid"], response)

    def sendMessage(self, payload, isBinary=False, fragmentSize=None, sync=False, doNotCompress=False):
        """
        Send a message over the WebSocket.
        :param payload: Message to send.
        """
        payload["msg"] = self.msg
        self.msg += 1
        payload = json.dumps(payload, sort_keys=True)
        self.logger.debug("Sent data: %s" % payload)
        payload = payload.encode("utf-8")
        super(DSAWebSocket, self).sendMessage(payload, isBinary, fragmentSize, sync, doNotCompress)



from setuptools import setup

setup(
    name="dslink",
    version="0.6.11",
    description="DSLink SDK for Python",
    url="http://github.com/IOT-DSA/sdk-dslink-python",
    author="Logan Gorence",
    author_email="l.gorence@dglogik.com",
    license="Apache 2.0",
    packages=[
        "dslink"
    ],
    install_requires=[
        "autobahn == 0.12.1",
        "pyelliptic == 1.5.7",
        "requests == 2.9.1",
        "zope.interface == 4.1.3"
    ],
    use_2to3=True
)
